#include <vector>
#include <numeric>

#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>
#include <unistd.h>

#if ESD
#include <esd.h>
#include <climits>

static int esd=-1;

void esdclean()
{
  if(esd>=0)
    esd_close(esd);
}

void esdinit()
{
  atexit(esdclean);

  esd_format_t format=ESD_BITS16|ESD_MONO|ESD_STREAM|ESD_MONITOR;

  esd=esd_monitor_stream(format, ESD_DEFAULT_RATE, getenv("ESPEAKER"), "LED Monitor");
  if(esd<0)
    {
      fprintf(stderr, "could not init ESD\n");
      exit(1);
    }

}

double getEsd()
{
  int16_t buf[ESD_BUF_SIZE];

  int r=read(esd, buf, ESD_BUF_SIZE);
  r/=sizeof(buf[0]);
  int max=0;
  for(int i=0; i<r; i++)
    {
      int b=buf[i];
      if(b>max)
	max=b;
      else
	{
	  b=-b;
	  if(b>max)
	    max=b;
	}
    }

  return static_cast<double>(max)/static_cast<double>(SHRT_MAX);
}

#endif

/**
   send text to syslog with priority
*/
void logger(const int priority, const char *text)
{
  openlog("showled", LOG_PID, LOG_DAEMON);
  syslog(priority, text);
  closelog();
}

/**
   this function makes a daemon process out of the current process.
   @return 0 on success. the current process is now a daemon
   -1 on error. the process should be exited, because an error occured while becoming a daemon
*/
int daemon()
{
  char trash[1024];

  // fork so the parent which invoked the program exists to the shell
  switch(fork())
    {
    case -1: 
      snprintf(trash, sizeof(trash), "fork() error: %s ; aborting...", strerror(errno));
      logger(LOG_ERR, trash); 
      return -1;
    case  0: break;
    default: exit(0);
    }

  // become process group leader of a new process group
  if(setsid()==-1)
    {
      logger(LOG_ERR, "failed setsid() ! aborting...");
      return -1;
    }

  // exit once again, so we are not the leader anymore
  switch(fork())
    {
    case -1: 
      snprintf(trash, sizeof(trash), "fork() error: %s ; aborting...", strerror(errno));
      logger(LOG_ERR, trash); 
      return -1;
    case  0: break;
    default: exit(0);
    }

  // change working directory to / to avoid blocking mounted filesystems
  if(chdir("/")==-1)
    {
      snprintf(trash, sizeof(trash), "failed chdir(\"/\"): %s ; aborting...", strerror(errno));
      logger(LOG_ERR, trash);
      return -1;
    }

  // set filemask
  umask(0);

  // close all files
  for(int i=sysconf(_SC_OPEN_MAX); i>=0; i--)
    close(i);

  // open stdin, stdout, stderr again to the famous /dev/null
  if(open("/dev/null", O_RDWR) < 0)
    {
      snprintf(trash, sizeof(trash), "failed to open stdin: %s ; aborting...", strerror(errno));
      logger(LOG_ERR, trash);
      return -1;
  }
  if(dup(0) < 0)
    {
      snprintf(trash, sizeof(trash), "failed to open stdout: %s ; aborting...", strerror(errno));
      logger(LOG_ERR, trash);
      return -1;
    }
  if(dup(0) < 0)
    {
      snprintf(trash, sizeof(trash), "failed to open stderr: %s ; aborting...", strerror(errno));
      logger(LOG_ERR, trash);
      return -1;
    }
  
  return 0;
}


/** returns the current usage of the CPU calculated from /proc/stat
    @return -1 on error, cpuUsage otherwise
*/
double getCPU() 
{
  static int prevuser=0, previdle=0;

  FILE *f=fopen("/proc/stat", "r");
  if(f==0)
    {
      logger(LOG_ERR, "could not open /proc/stat");
      exit(1);
    }

  char cpu[10];
  int user, nice, system, idle;
  if( fscanf(f,"%s %i %i %i %i", cpu, &user, &nice, &system, &idle) != 5 )
    {
      fclose(f);
      logger(LOG_ERR, "could not read from /proc/stat");
      exit(1);
    }

  fclose(f);

  // sum together user+nice+system
  user+=nice;
  user+=system;

  const int u=user-prevuser; prevuser=user;
  const int i=idle-previdle; previdle=idle;

  // no idle time, thus cpu was fully used
  if(i==0)
    return 1;

  if(u==0 || (u+i)==0)
    return 0;

  return static_cast<double>(u)/static_cast<double>(u+i);
}

/** sleep for usec */
void sl(int usec)
{
  struct timeval tv;
  tv.tv_sec = usec/1000000;
  tv.tv_usec = usec%1000000;

  if(select(1, NULL, NULL, NULL, &tv) < 0)
    {
      perror("could not select");
      exit(1);
    }
}

int ledfh=-1;

void esistaus()
{
  if(ledfh>=0)
    {
      uint32_t c=0;
      write(ledfh, &c, sizeof(c));
      close(ledfh);
    }
}

int main(int argc, char **argv)
{
  atexit(esistaus);

  double (*get)() = getCPU;

  if(argc>1)
    {
#if ESD
      if(!strcmp(argv[1], "esd"))
	{
	  esdinit();
	  get=getEsd;
	}
#endif
      if(!strcmp(argv[1], "-d"))
	if(daemon() < 0)
	  {
	    fprintf(stderr, "could not become a daemon. aborting...");
	    return 1;
	  }
    }

  ledfh=open("/dev/led", O_WRONLY);
  if(ledfh<0)
    ledfh=open("/dev/misc/led", O_WRONLY);
  if(ledfh<0)
    {
      fprintf(stderr, "could not open /dev/led or /dev/misc/led\n");
      exit(1);
    }

  int vi=0;
  std::vector<double> v(10, 0.0);
  while(true)
    {
      v[vi++]=get(); vi%=v.size();

      const double d=std::accumulate(v.begin(), v.end(), 0.0)/v.size();

      uint32_t c=static_cast<uint32_t>(d*32.0);
      if(c>=31) 
	c=~0;
      else
	c=(1<<c)-1;
      
      write(ledfh, &c, sizeof(c));

      sl(50000);
    }
  return 0;
}
